import json
from typing import Optional

from rich.table import Table

from ..authclient import AuthClient
from ..config import settings
from ..interface import BaseInterface
from ..interface import PrintInterface
from ..interface import RichConsoleInterface
from ..interface import RichJsonInterface
from ..response import check_response


def post_dataset(
    client: AuthClient,
    *,
    project_id: int,
    dataset_name: str,
    metadata: Optional[str] = None,
    type: Optional[str] = None,
    batch: bool = False,
    make_read_only: bool = False,
) -> RichJsonInterface:
    """
    Arguments:
        project_id: ID of project to add the new dataset to
        dataset_name: Name of new dataset
        metadata: Path to file containing dataset metadata in JSON format.
        type: Dataset type.
        make_read_only: Make the new dataset read-only.
    """
    if metadata is None:
        meta = {}
    else:
        with open(metadata, "r") as f:
            meta = json.load(f)

    dataset = dict(name=dataset_name, meta=meta, read_only=make_read_only)
    if type:
        dataset["type"] = type

    res = client.post(
        f"{settings.BASE_URL}/project/{project_id}/dataset/",
        json=dataset,
    )
    new_dataset = check_response(res, expected_status_code=201)
    if batch:
        return PrintInterface(retcode=0, data=new_dataset["id"])
    else:
        return RichJsonInterface(retcode=0, data=new_dataset)


def post_resource(
    client: AuthClient,
    *,
    project_id: int,
    dataset_id: int,
    path: str,
    batch: bool = False,
) -> BaseInterface:

    res = client.post(
        (
            f"{settings.BASE_URL}/project/{project_id}/"
            f"dataset/{dataset_id}/resource/"
        ),
        json=dict(path=path),
    )
    new_resource = check_response(res, expected_status_code=201)
    if batch:
        return PrintInterface(retcode=0, data=new_resource["id"])
    else:
        return RichJsonInterface(retcode=0, data=new_resource)


def delete_resource(
    client: AuthClient,
    *,
    project_id: int,
    dataset_id: int,
    resource_id: int,
) -> BaseInterface:
    res = client.delete(
        (
            f"{settings.BASE_URL}/project/{project_id}/"
            f"dataset/{dataset_id}/resource/{resource_id}/"
        )
    )
    check_response(res, expected_status_code=204)
    return PrintInterface(retcode=0, data="")


def patch_dataset(
    client: AuthClient,
    *,
    project_id: int,
    dataset_id: int,
    new_name: Optional[str] = None,
    new_type: Optional[str] = None,
    meta_file: Optional[str] = None,
    make_read_only: bool = False,
    remove_read_only: bool = False,
) -> BaseInterface:

    # Prepare payload
    dataset_update = {}
    if new_name:
        dataset_update["name"] = new_name
    if new_type:
        dataset_update["type"] = new_type
    if meta_file:
        with open(meta_file, "r") as f:
            meta = json.load(f)
        dataset_update.update(meta=meta)
    if make_read_only:
        dataset_update["read_only"] = True
    if remove_read_only:
        dataset_update["read_only"] = False

    if not dataset_update:
        return PrintInterface(retcode=1, data="Nothing to update")

    res = client.patch(
        (
            f"{settings.BASE_URL}/project/{project_id}/"
            f"dataset/{dataset_id}/"
        ),
        json=dataset_update,
    )
    data = check_response(res, expected_status_code=200)
    return RichJsonInterface(retcode=0, data=data)


def get_dataset(
    client: AuthClient, *, project_id: int, dataset_id: int
) -> BaseInterface:
    res = client.get(
        f"{settings.BASE_URL}/project/{project_id}/dataset/{dataset_id}/"
    )
    from rich.console import Group

    dataset = check_response(res, expected_status_code=200)

    table = Table(title="Dataset")
    table.add_column("ID", style="cyan", no_wrap=True)
    table.add_column("Name", justify="right", style="green")
    table.add_column("Type", style="white")
    table.add_column("Meta", justify="center")
    table.add_column("Read only", justify="center")

    table.add_row(
        str(dataset["id"]),
        dataset["name"],
        dataset["type"],
        json.dumps(dataset["meta"], indent=2),
        "✅" if dataset["read_only"] else "❌",
    )
    table_res = Table(title="Resources")
    table_res.add_column("Path", justify="center", style="yellow")
    table_res.add_column("ID", justify="center", style="yellow")
    table_res.add_column("Dataset ID", justify="center", style="yellow")
    for r in dataset["resource_list"]:
        table_res.add_row(r["path"], str(r["id"]), str(r["dataset_id"]))
    group = Group(table, table_res)
    return RichConsoleInterface(retcode=0, data=group)


def delete_dataset(
    client: AuthClient, *, project_id: int, dataset_id: int
) -> PrintInterface:

    res = client.delete(
        f"{settings.BASE_URL}/project/{project_id}/dataset/{dataset_id}/"
    )
    check_response(res, expected_status_code=204)
    return PrintInterface(retcode=0, data="")


def get_dataset_history(
    client: AuthClient, *, project_id: int, dataset_id: int
) -> BaseInterface:
    res = client.get(
        f"{settings.BASE_URL}/project/{project_id}/dataset/{dataset_id}/"
        "export_history/"
    )
    history_workflow = check_response(res, expected_status_code=200)
    return RichJsonInterface(retcode=0, data=history_workflow)


def get_dataset_status(
    client: AuthClient, *, project_id: int, dataset_id: int
) -> BaseInterface:
    res = client.get(
        f"{settings.BASE_URL}/project/{project_id}/dataset/{dataset_id}/"
        "status/"
    )
    dataset_status = check_response(res, expected_status_code=200)
    return RichJsonInterface(retcode=0, data=dataset_status)



import logging
from pathlib import Path

import jwt
from httpx import Client
from jwt.exceptions import ExpiredSignatureError

from .config import settings


def debug_request(verb: str, url: str, **kwargs):
    body = kwargs.get("json")
    log = f"\nFractal Client sending HTTP request to:\n    {verb} {url}"
    if body is not None:
        log += "\nRequest body:\n" + "\n".join(
            [f"    {k}: {v}" for k, v in body.items()]
        )
    logging.debug(log)


class AuthenticationError(ValueError):
    pass


class AuthToken:
    def __init__(
        self,
        client: Client,
        username: str,
        password: str,
    ):
        self.client = client
        self.username = username
        self.password = password

        try:
            with open(f"{settings.FRACTAL_CACHE_PATH}/session", "r") as f:
                self.token = f.read()
        except FileNotFoundError:
            pass

    def _get_fresh_token(self):
        data = dict(
            username=self.username,
            password=self.password,
        )
        res = self.client.post(
            f"{settings.FRACTAL_SERVER}/auth/token/login/", data=data
        )
        if res.status_code != 200:
            data = res.text
            raise AuthenticationError(
                f"Error at {res.request.url}.\n"
                f"Status code: {res.status_code}.\n"
                f"Response data: {data}.\n"
            )
        raw_token = res.json()
        self.token = raw_token["access_token"]

        # Create cache folder, if needed
        cache_dir = Path(f"{settings.FRACTAL_CACHE_PATH}")
        cache_dir.mkdir(parents=True, exist_ok=True)

        # Write token in cache_file
        cache_file = cache_dir / "session"
        with cache_file.open("w") as f:
            f.write(self.token)

    @property
    def expired(self):
        try:
            jwt.decode(
                jwt=self.token,
                options={
                    "verify_signature": False,
                    "verify_exp": True,
                },
            )
            return False
        except AttributeError:
            return True
        except ExpiredSignatureError:
            return True

    def header(self):
        token = self.__call__()
        return dict(Authorization=f"Bearer {token}")

    def __call__(self):
        if self.expired:
            self._get_fresh_token()
        return self.token


class AuthClient:
    def __init__(
        self,
        username: str,
        password: str,
    ):
        self.auth = None
        self.client = None
        self.username = username
        self.password = password

    def __enter__(self):
        logging.getLogger("httpx").setLevel(logging.WARNING)
        self.client = Client()
        self.auth = AuthToken(
            client=self.client,
            username=self.username,
            password=self.password,
        )
        return self

    def __exit__(self, *args):
        self.client.close()

    def get(self, *args, **kwargs):
        debug_request("GET", args[0], **kwargs)
        return self.client.get(headers=self.auth.header(), *args, **kwargs)

    def post(self, *args, **kwargs):
        debug_request("POST", args[0], **kwargs)
        return self.client.post(headers=self.auth.header(), *args, **kwargs)

    def patch(self, *args, **kwargs):
        debug_request("PATCH", args[0], **kwargs)
        return self.client.patch(headers=self.auth.header(), *args, **kwargs)

    def delete(self, *args, **kwargs):
        debug_request("DELETE", args[0], **kwargs)
        return self.client.delete(headers=self.auth.header(), *args, **kwargs)



from typing import Optional
from typing import Union

from rich.table import Table

from ..authclient import AuthClient
from ..config import settings
from ..interface import BaseInterface
from ..interface import PrintInterface
from ..interface import RichConsoleInterface
from ..interface import RichJsonInterface
from ..response import check_response


def post_project(
    client: AuthClient,
    *,
    name: str,
    batch: bool = False,
) -> BaseInterface:
    # Prepare a ProjectCreate request body
    payload = dict(name=name)
    # Send API request
    res = client.post(f"{settings.BASE_URL}/project/", json=payload)
    project = check_response(res, expected_status_code=201)
    if batch:
        return PrintInterface(retcode=0, data=project["id"])
    else:
        return RichJsonInterface(retcode=0, data=project)


def get_project_list(client: AuthClient) -> RichConsoleInterface:

    res = client.get(f"{settings.BASE_URL}/project/")
    projects = check_response(res, expected_status_code=200)

    table = Table(title="Project List")
    table.add_column("ID", style="cyan", no_wrap=True)
    table.add_column("Name", style="magenta")
    table.add_column("Read only", justify="center")

    for p in projects:
        # Map p.read_only (True/False) to read_only_icon (✅/❌)
        if p["read_only"]:
            read_only_icon = "✅"
        else:
            read_only_icon = "❌"

        table.add_row(
            str(p["id"]),
            p["name"],
            read_only_icon,
        )

    return RichConsoleInterface(retcode=0, data=table)


def get_project(client: AuthClient, *, project_id: int) -> RichJsonInterface:
    res = client.get(f"{settings.BASE_URL}/project/{project_id}/")
    project = check_response(res, expected_status_code=200)
    return RichJsonInterface(retcode=0, data=project)


def delete_project(client: AuthClient, *, project_id: int) -> PrintInterface:

    res = client.delete(f"{settings.BASE_URL}/project/{project_id}/")
    check_response(res, expected_status_code=204)
    return PrintInterface(retcode=0, data="")


def patch_project(
    client: AuthClient,
    *,
    project_id: int,
    new_name: Optional[str] = None,
    make_read_only: bool = False,
    remove_read_only: bool = False,
) -> Union[RichJsonInterface, PrintInterface]:
    project_update = {}
    if new_name:
        project_update["name"] = new_name
    if make_read_only:
        project_update["read_only"] = True
    if remove_read_only:
        project_update["read_only"] = False

    if not project_update:
        return PrintInterface(retcode=1, data="Nothing to update")

    res = client.patch(
        f"{settings.BASE_URL}/project/{project_id}/", json=project_update
    )
    new_project = check_response(res, expected_status_code=200)
    return RichJsonInterface(retcode=0, data=new_project)



import json
import logging
import sys
from pathlib import Path
from typing import Optional

from ..authclient import AuthClient
from ..config import settings
from ..interface import BaseInterface
from ..interface import PrintInterface
from ..interface import RichJsonInterface
from ..response import check_response
from ._aux_task_caching import FractalCacheError
from ._aux_task_caching import get_task_id_from_cache


def post_workflow(
    client: AuthClient, *, name: str, project_id: int, batch: bool = False
) -> BaseInterface:
    workflow = dict(
        name=name,
        project_id=project_id,
    )
    res = client.post(
        f"{settings.BASE_URL}/project/{project_id}/workflow/",
        json=workflow,
    )
    workflow = check_response(res, expected_status_code=201)
    if batch:
        return PrintInterface(retcode=0, data=workflow["id"])
    else:
        return RichJsonInterface(retcode=0, data=workflow)


def get_workflow_list(
    client: AuthClient, *, project_id: int, batch: bool = False
) -> RichJsonInterface:

    res = client.get(f"{settings.BASE_URL}/project/{project_id}/workflow/")
    workflow_list = check_response(res, expected_status_code=200)
    return RichJsonInterface(retcode=0, data=workflow_list)


def delete_workflow(
    client: AuthClient, *, project_id: int, workflow_id: int
) -> BaseInterface:
    res = client.delete(
        f"{settings.BASE_URL}/project/{project_id}/workflow/{workflow_id}/"
    )
    check_response(res, expected_status_code=204)
    return PrintInterface(retcode=0, data="")


def get_workflow(
    client: AuthClient, *, project_id: int, workflow_id: int
) -> RichJsonInterface:
    res = client.get(
        f"{settings.BASE_URL}/project/{project_id}/workflow/{workflow_id}/"
    )
    workflow = check_response(res, expected_status_code=200)
    return RichJsonInterface(retcode=0, data=workflow)


def post_workflowtask(
    client: AuthClient,
    *,
    project_id: int,
    workflow_id: int,
    task_id: Optional[int] = None,
    task_name: Optional[str] = None,
    task_version: Optional[str] = None,
    batch: bool = False,
    order: Optional[int] = None,
    args_file: Optional[str] = None,
    meta_file: Optional[str] = None,
) -> RichJsonInterface:

    if task_id:
        if task_version:
            logging.error(
                "Too many arguments: cannot provide both "
                "`task_id` and `task_version`."
            )
            sys.exit(1)
    else:
        try:
            task_id = get_task_id_from_cache(
                client=client, task_name=task_name, version=task_version
            )
        except FractalCacheError as e:
            print(e)
            sys.exit(1)

    if order is None:
        workflow_task = dict()
    else:
        workflow_task = dict(order=order)
    if args_file:
        with Path(args_file).open("r") as f:
            args = json.load(f)
            workflow_task["args"] = args
    if meta_file:
        with Path(meta_file).open("r") as f:
            meta = json.load(f)
            workflow_task["meta"] = meta

    res = client.post(
        (
            f"{settings.BASE_URL}/project/{project_id}/"
            f"workflow/{workflow_id}/wftask/"
            f"?{task_id=}"
        ),
        json=workflow_task,
    )
    workflow_task = check_response(res, expected_status_code=201)

    if batch:
        return PrintInterface(retcode=0, data=str(workflow_task["id"]))
    else:
        return RichJsonInterface(retcode=0, data=workflow_task)


def patch_workflowtask(
    client: AuthClient,
    *,
    project_id: int,
    workflow_id: int,
    workflow_task_id: int,
    args_file: Optional[str] = None,
    meta_file: Optional[str] = None,
) -> RichJsonInterface:

    # Check that at least one of args_file or meta_file was given (note: it
    # would be reasonable to check it in the parser, but we are not aware of a
    # method within argparse).
    if not (args_file or meta_file):
        raise ValueError(
            "At least one of {args_file,meta_file} arguments is required"
        )

    # Combine args/meta payload
    payload = {}
    if args_file:
        with Path(args_file).open("r") as f:
            args = json.load(f)
            payload["args"] = args
    if meta_file:
        with Path(meta_file).open("r") as f:
            meta = json.load(f)
            payload["meta"] = meta

    res = client.patch(
        (
            f"{settings.BASE_URL}/project/{project_id}/"
            f"workflow/{workflow_id}/wftask/{workflow_task_id}/"
        ),
        json=payload,
    )
    workflow_task = check_response(res, expected_status_code=200)

    return RichJsonInterface(retcode=0, data=workflow_task)


def delete_workflowtask(
    client: AuthClient,
    *,
    project_id: int,
    workflow_id: int,
    workflow_task_id: int,
) -> BaseInterface:

    res = client.delete(
        f"{settings.BASE_URL}/project/{project_id}/"
        f"workflow/{workflow_id}/wftask/{workflow_task_id}/"
    )
    check_response(res, expected_status_code=204)
    return PrintInterface(retcode=0, data="")


def patch_workflow(
    client: AuthClient,
    *,
    project_id: int,
    workflow_id: int,
    new_name: str,
) -> BaseInterface:

    workflow_update = dict(name=new_name)

    res = client.patch(
        f"{settings.BASE_URL}/project/{project_id}/workflow/{workflow_id}/",
        json=workflow_update,
    )
    new_workflow = check_response(res, expected_status_code=200)
    return RichJsonInterface(retcode=0, data=new_workflow)


def workflow_apply(
    client: AuthClient,
    *,
    project_id: int,
    workflow_id: int,
    input_dataset_id: int,
    output_dataset_id: int,
    first_task_index: Optional[int] = None,
    last_task_index: Optional[int] = None,
    worker_init: Optional[str] = None,
    batch: bool = False,
) -> BaseInterface:
    apply_wf_create = dict(
        workflow_id=workflow_id,
        input_dataset_id=input_dataset_id,
        output_dataset_id=output_dataset_id,
    )
    # Prepare ApplyWorkflowCreate object, without None attributes
    if worker_init is not None:
        apply_wf_create["worker_init"] = worker_init
    if first_task_index is not None:
        apply_wf_create["first_task_index"] = first_task_index
    if last_task_index is not None:
        apply_wf_create["last_task_index"] = last_task_index

    # Prepare query parameters
    query_parameters = (
        f"input_dataset_id={input_dataset_id}"
        f"&output_dataset_id={output_dataset_id}"
    )

    res = client.post(
        (
            f"{settings.BASE_URL}/project/{project_id}/workflow/{workflow_id}/"
            f"apply/?{query_parameters}"
        ),
        json=apply_wf_create,
    )
    apply_wf_read = check_response(res, expected_status_code=202)

    if batch:
        return PrintInterface(retcode=0, data=apply_wf_read["id"])
    else:
        return RichJsonInterface(retcode=0, data=apply_wf_read)


def workflow_import(
    client: AuthClient, *, project_id: int, json_file: str, batch: bool = False
) -> BaseInterface:
    with Path(json_file).open("r") as f:
        workflow = json.load(f)

    res = client.post(
        f"{settings.BASE_URL}/project/{project_id}/workflow/import/",
        json=workflow,
    )
    wf_read = check_response(res, expected_status_code=201)

    warnings = [
        workflow_task["task"]["source"]
        for workflow_task in wf_read["task_list"]
        if workflow_task["task"]["owner"]
    ]
    if warnings:
        sources_str = ", ".join([f"'{s}'" for s in warnings])
        logging.warning(
            "This workflow includes custom tasks (the ones with sources: "
            f"{sources_str}), which are not meant to be portable; "
            "importing this workflow may not work as expected."
        )

    if batch:
        datastr = f"{wf_read['id']}"
        for wftask in wf_read["task_list"]:
            datastr += f" {wftask['id']}"
        return PrintInterface(retcode=0, data=datastr)
    else:
        return RichJsonInterface(retcode=0, data=wf_read)


def workflow_export(
    client: AuthClient,
    *,
    project_id: int,
    workflow_id: int,
    json_file: str,
) -> BaseInterface:
    res = client.get(
        (
            f"{settings.BASE_URL}/project/{project_id}/"
            f"workflow/{workflow_id}/export/"
        ),
    )
    workflow = check_response(res, expected_status_code=200)

    warnings = [
        workflow_task["task"]["source"]
        for workflow_task in workflow["task_list"]
        if not workflow_task["task"]["source"].startswith(
            ("pip_local:", "pip_remote:")
        )
    ]
    if warnings:
        sources_str = ", ".join([f"'{s}'" for s in warnings])
        logging.warning(
            "This workflow includes custom tasks (the ones with sources: "
            f"{sources_str}), which are not meant to be portable; "
            "re-importing this workflow may not work as expected."
        )

    with Path(json_file).open("w") as f:
        json.dump(workflow, f, indent=2)
    return PrintInterface(
        retcode=0, data=f"Workflow {workflow_id} exported at {json_file}"
    )
